package main

import (
	"fmt"
	"net/http"
	"io/ioutil"
	"crypto/tls"
	"time"
	"encoding/json"
	"bytes"
	"net/url"
	"os"
	"io"
)

const Site24x7TokensFile = "site24x7_tokens.json"
const HTTP_GET = "GET"
const HTTP_POST = "POST"

type Site24x7OAuthTokens struct {
	CLIENT_ID string
	CLIENT_SECRET string
	GENERATED_CODE string
	REFRESH_TOKEN string
	ACCESS_TOKEN string
	TOKEN_GENERATION_TIME int64
	EXPIRES_IN_SEC float64
}

type Site24x7OAuthValidator struct{}

type Site24x7Connector struct{
	Connector
}

type ZohoAccountsConnector struct{
	Connector
	TokenApi string
}

type Connector struct{
	Domain string
	Timeout int64
}

var OAuthTokens Site24x7OAuthTokens
var OAuthValidator Site24x7OAuthValidator
var BaseConnector Connector
var Site247Connector Site24x7Connector
var AccountsConnector ZohoAccountsConnector

func init(){
	Site247Connector.init()
	AccountsConnector.init()
	isTokPresent := OAuthValidator.isTokenPresent()
	if(isTokPresent){
		if(!OAuthValidator.isValidToken()){
			AccountsConnector.getAccessTokenFromRefreshToken()
		}
	}
}


//Connector methods

func (connector *Connector) getClient() http.Client{
	trTls11 := &http.Transport{
		DisableKeepAlives:true,
		TLSClientConfig: &tls.Config{
			MaxVersion: tls.VersionTLS13,
			InsecureSkipVerify: true,
		},
	}
	client := http.Client{
		Transport: trTls11,
		Timeout:   connector.getTimeout(),
	}
	return client
}

func (connector *Connector) getRequest(requestType string, url string, body io.Reader) (*http.Request, error){
	return http.NewRequest(requestType, url, body)
}

func (connector *Connector) getTimeout() time.Duration{
	return time.Duration(connector.Timeout * int64(time.Second))
}

func (connector *Connector) setDefaultHeaders(request *http.Request){
	request.Header.Set("Content-Type","application/json;charset=UTF-8")
	request.Header.Set("Accept","application/json; version=2.0")
	request.Header.Set("Authorization","Zoho-oauthtoken "+OAuthTokens.ACCESS_TOKEN)
}

//ZohoAccountsConnector methods

func (accountsConnector *ZohoAccountsConnector) init() {
	accountsConnector.Domain = "https://accounts.zoho.com"
	accountsConnector.TokenApi = "/oauth/v2/token"
}

func (accountsConnector *ZohoAccountsConnector) getUrl(urlValues url.Values) string{
	baseAccurl := fmt.Sprintf("%s%s", accountsConnector.Domain, accountsConnector.TokenApi)
	urlToReturn := fmt.Sprintf("%s?%s", baseAccurl, urlValues.Encode())
	return urlToReturn
}

func (accountsConnector *ZohoAccountsConnector) getUrlValuesForAccessTokenFromRefreshToken() url.Values{
	urlValues := url.Values{}
	urlValues.Set("client_id", OAuthTokens.CLIENT_ID)
	urlValues.Set("client_secret", OAuthTokens.CLIENT_SECRET)
	urlValues.Set("refresh_token", OAuthTokens.REFRESH_TOKEN)
	urlValues.Set("grant_type", "refresh_token")
	return urlValues
}

func (accountsConnector *ZohoAccountsConnector) getUrlValuesForAccessTokenFromGeneratedCode() url.Values{
	urlValues := url.Values{}
	urlValues.Set("client_id", OAuthTokens.CLIENT_ID)
	urlValues.Set("client_secret", OAuthTokens.CLIENT_SECRET)
	urlValues.Set("code", OAuthTokens.GENERATED_CODE)
	urlValues.Set("grant_type", "authorization_code")
	return urlValues
}

func (accountsConnector *ZohoAccountsConnector) setDefaultHeaders(request *http.Request){
	request.Header.Set("Accept","application/x-www-form-urlencoded")
}

func (accountsConn ZohoAccountsConnector) getAccessTokenFromRefreshToken(){
	fmt.Println("====================== GENERATING ACCESS TOKEN FROM REFRESH TOKEN =========================")
	var result map[string]interface{}
	urlToPost := accountsConn.getUrl(accountsConn.getUrlValuesForAccessTokenFromRefreshToken())
	req, _ := accountsConn.getRequest(HTTP_POST, urlToPost, nil)
	Site247Connector.setDefaultHeaders(req)
	client := accountsConn.getClient()
	
	res, err := client.Do(req)
	
	if(err == nil){
		fmt.Println("Response Status:", res.Status)
		defer res.Body.Close()
		json.NewDecoder(res.Body).Decode(&result)
		OAuthTokens.ACCESS_TOKEN = result["access_token"].(string)
		OAuthTokens.EXPIRES_IN_SEC = result["expires_in_sec"].(float64)
		if(OAuthTokens.ACCESS_TOKEN == "" || OAuthTokens.EXPIRES_IN_SEC == 0){
			fmt.Println("Error while fetching access token from refresh token :",err)
			os.Exit(1)	
		}else{
			OAuthTokens.TOKEN_GENERATION_TIME = NowAsUnixMilli()
			OAuthTokens.persist()	
		}	
	}else{
		fmt.Println("Error while fetching access token from refresh token :",err)
		os.Exit(1)
	}
}

func (accountsConn ZohoAccountsConnector) setAccessTokenFromCode(){
	fmt.Println("====================== GENERATING ACCESS TOKEN AND REFRESH TOKEN FROM GENERATED CODE =========================")
	var result map[string]interface{}
	urlToPost := accountsConn.getUrl(accountsConn.getUrlValuesForAccessTokenFromGeneratedCode())
	req, _ := accountsConn.getRequest(HTTP_POST, urlToPost, nil)
	Site247Connector.setDefaultHeaders(req)
	client := accountsConn.getClient()
	res, err := client.Do(req)
	
	if(err == nil){
		fmt.Println("Response Status:", res.Status)
		defer res.Body.Close()
		json.NewDecoder(res.Body).Decode(&result)
		OAuthTokens.ACCESS_TOKEN = result["access_token"].(string)
		OAuthTokens.REFRESH_TOKEN = result["refresh_token"].(string)
		OAuthTokens.EXPIRES_IN_SEC = result["expires_in_sec"].(float64)
		if(OAuthTokens.REFRESH_TOKEN == "" || OAuthTokens.ACCESS_TOKEN == "" || OAuthTokens.EXPIRES_IN_SEC == 0){
			fmt.Println("Error while fetching access token and refresh token from generated code :",err)
			os.Exit(1)	
		}else{
			OAuthTokens.TOKEN_GENERATION_TIME = NowAsUnixMilli()
			OAuthTokens.persist()	
		}	
	}else{
		fmt.Println("Error while fetching access token and refresh token from generated code :",err)
		os.Exit(1)
	}
}

//Site24x7Connector methods

func (s247Connector *Site24x7Connector) init() {
	s247Connector.Domain = "https://www.site24x7.com"
}

func (s247Connector *Site24x7Connector) getUrl(api string) string{
	urlToReturn := fmt.Sprintf("%s%s", s247Connector.Domain, api)
	return urlToReturn
}


//Site24x7OAuthTokens methods

func (oauthCred *Site24x7OAuthTokens) persist(){
	file, _ := json.MarshalIndent(OAuthTokens, "", " ")
	_ = ioutil.WriteFile(Site24x7TokensFile, file, 0644)
}

func (oauthCred *Site24x7OAuthTokens) loadTokens(){
	jsonFile, err := os.Open(Site24x7TokensFile)
	if err != nil {
	    fmt.Println("Error while loading tokens : ",err)
	}
	defer jsonFile.Close()
	byteValue, _ := ioutil.ReadAll(jsonFile)
	json.Unmarshal(byteValue, &OAuthTokens)
}

// Site24x7OAuthValidator methods

func (validator *Site24x7OAuthValidator) isTokenPresent() bool{
	isTokenPresent := false
	if(fileExists(Site24x7TokensFile)){
		OAuthTokens.loadTokens()
		//fmt.Println("OAuthTokens.ACCESS_TOKEN : ",OAuthTokens.ACCESS_TOKEN)
		if OAuthTokens.ACCESS_TOKEN != ""{
			isTokenPresent = true
		}else if(OAuthTokens.CLIENT_ID != "" && OAuthTokens.CLIENT_SECRET != "" && OAuthTokens.GENERATED_CODE != ""){
			AccountsConnector.setAccessTokenFromCode()
			
		}else{
			fmt.Println("Please update CLIENT_ID, CLIENT_SECRET and GENERATED_CODE in the file",Site24x7TokensFile)
			os.Exit(1)
		}
	}else{
		OAuthTokens.persist()
		fmt.Println("Please update CLIENT_ID, CLIENT_SECRET and GENERATED_CODE in the file",Site24x7TokensFile)
		os.Exit(1)
	}
	fmt.Println("Is Access Token Present : ",isTokenPresent)
	return isTokenPresent
}

func (validator *Site24x7OAuthValidator) isValidToken() bool{
	if((NowAsUnixMilli() - OAuthTokens.TOKEN_GENERATION_TIME) > int64(OAuthTokens.EXPIRES_IN_SEC)*1000){
		return false
	}else{
		return true
	}
}

// Util functions

func fileExists(filename string) bool {
    info, err := os.Stat(filename)
    if os.IsNotExist(err) {
        return false
    }
    return !info.IsDir()
}

func NowAsUnixMilli() int64 {
    return time.Now().UnixNano() / 1e6
}

func sampleGet(api string){
	urlToGet := Site247Connector.getUrl(api)
	fmt.Println("====================== Sending GET request for the url : ",urlToGet," =========================")
	req, _ := Site247Connector.getRequest(HTTP_GET, urlToGet, nil)
	Site247Connector.setDefaultHeaders(req)
	client := Site247Connector.getClient()
	res, err := client.Do(req)
	if res != nil{
		defer res.Body.Close()
		body, _ := ioutil.ReadAll(res.Body)
		fmt.Println("Response Status Code :", res.Status)
		fmt.Println("Response : ",string(body))
	} else{
		fmt.Println("Error : ",err)
	}
}

func samplePost(api string){
	urlToPost := Site247Connector.getUrl(api)
	fmt.Println("====================== Sending POST request for the url : ",urlToPost," =========================")
	params, _  := json.Marshal((map[string]string{
		"name" : "Site24x7-Slack Integration",
		"url" : "https://hooks.slack.com/services/123123/2112312",
		"sender_name" : "Site24x7",
		"selection_type" : "0",
		"title" : "$MONITOR_NAME is $STATUS",
	}))
	req, _ := Site247Connector.getRequest(HTTP_POST, urlToPost, bytes.NewBuffer(params))
	Site247Connector.setDefaultHeaders(req)
	client := Site247Connector.getClient()
	
	res, err := client.Do(req)
	if res != nil{
		defer res.Body.Close()
		
		body, _ := ioutil.ReadAll(res.Body)
		fmt.Println("Response Status Code :", res.Status)
		fmt.Println("Response : ",string(body))
	} else{
		fmt.Println("Error : ",err)
	}
	
	
}

func main() {
	fmt.Println("Is valid access token : ",OAuthValidator.isValidToken())
	//Uncomment the below line for generating access token from refresh token forcefully
	//AccountsConnector.getAccessTokenFromRefreshToken()
	if(OAuthValidator.isValidToken()){
		sampleGet("/api/it_automation")
		//samplePost("/api/integration/slack")
	}
	
}